/*
  Copyright (C) 2005 Jack Lloyd <lloyd@randombit.net>
*/

#include <cutlass/filters.hpp>
#include <botan/rng.h>
#include <botan/lookup.h>
using namespace Botan;

static void dump(const char* name, const SecureVector<byte>& x)
   {
   printf("%s (%d) = ", name, x.size());
   for(int j = 0; j != x.size(); j++)
      printf("%02X", x[j]);
   printf("\n");
   }

Cutlass_Encryptor::Cutlass_Encryptor(const std::string& cipher_algo,
                                     const std::string& hmac_hash,
                                     const SymmetricKey& cipher_key,
                                     const SymmetricKey& mac_key,
                                     u32bit mtu_val)
   {
   if(!have_block_cipher(cipher_algo))
      throw Invalid_Argument("Unavailable cipher: " + cipher_algo);
   if(!have_hash(hmac_hash))
      throw Invalid_Argument("Unavailable hash: " + hmac_hash);

   InitializationVector dummy(block_size_of(cipher_algo));

#if 0
   /* MAC of ciphertext */
   crypto_layer.append(
      cipher = get_cipher(cipher_algo + "/CTR-BE", cipher_key, dummy, ENCRYPTION)
      );
   crypto_layer.append(
      new Fork(
         0, /* pass through to output buffer */
         mac = new MAC_Filter("HMAC(" + hmac_hash + ")", mac_key)
         )
      );
#else
   /* MAC of plaintext */
   crypto_layer.append(
      new Fork(
         cipher = get_cipher(cipher_algo + "/CTR-BE", cipher_key, dummy, ENCRYPTION),
         mac = new MAC_Filter("HMAC(" + hmac_hash + ")", mac_key)
         )
      );
#endif

   mtu = mtu_val;
   seq_no = 0;
   iv_len = block_size_of(cipher_algo);
   }

void Cutlass_Encryptor::start_msg()
   {
   msg_buf.destroy();
   }

void Cutlass_Encryptor::write(const byte data[], u32bit len)
   {
   msg_buf.append(data, len);
   }

void Cutlass_Encryptor::end_msg()
   {
   if(msg_buf.size() < 2)
      throw Invalid_Argument("Cutlass_Encryptor: No type/channel id provided");

   InitializationVector iv(iv_len);
   cipher->set_iv(iv);
   send(iv.bits_of());

   crypto_layer.start_msg();

#if 0
   mac->write(iv.begin(), iv.length());
   for(u32bit j = 0; j != 4; j++)
      crypto_layer.write(get_byte(j, seq_no));
   seq_no++;
#endif

   u16bit msg_size = msg_buf.size() - 2;
   // FIXME: MTU check!

#if 0
   crypto_layer.write(get_byte(0, msg_size));
   crypto_layer.write(get_byte(1, msg_size));

   crypto_layer.write(msg_buf);
#else
   crypto_layer.write(msg_buf[0]);
   crypto_layer.write(msg_buf[1]);
   crypto_layer.write(get_byte(0, msg_size));
   crypto_layer.write(get_byte(1, msg_size));

   crypto_layer.write(msg_buf + 2, msg_buf.size() - 2);
#endif

#if 0
   SecureVector<byte> chaff(mtu - msg_size);
   Global_RNG::randomize(chaff, chaff.size(), Nonce);
   crypto_layer.write(chaff);
#endif

   crypto_layer.end_msg();

   send(crypto_layer.read_all(crypto_layer.message_count()-2)); /* cipher */
   send(crypto_layer.read_all(crypto_layer.message_count()-1)); /* MAC */
   }

Cutlass_Decryptor::Cutlass_Decryptor(const std::string& cipher_algo,
                                     const std::string& hmac_hash,
                                     const SymmetricKey& cipher_key,
                                     const SymmetricKey& mac_key)
   {
   if(!have_block_cipher(cipher_algo))
      throw Invalid_Argument("Unavailable cipher: " + cipher_algo);
   if(!have_hash(hmac_hash))
      throw Invalid_Argument("Unavailable hash: " + hmac_hash);

   InitializationVector dummy(block_size_of(cipher_algo));

#if 0
   /* MAC of ciphertext */
   crypto_layer.append(
      new Fork(
         cipher = get_cipher(cipher_algo + "/CTR-BE", cipher_key, dummy, DECRYPTION), 
         mac = new MAC_Filter("HMAC(" + hmac_hash + ")", mac_key)
         )
      );
#else
   /* MAC of plaintext */
   crypto_layer.append(
      cipher = get_cipher(cipher_algo + "/CTR-BE", cipher_key, dummy, DECRYPTION)
      );
   crypto_layer.append(
      new Fork(
         0,
         mac = new MAC_Filter("HMAC(" + hmac_hash + ")", mac_key)
         )
      );
#endif

   iv_len = block_size_of(cipher_algo);
   mac_len = output_length_of(hmac_hash);
   }

void Cutlass_Decryptor::start_msg()
   {
   msg_buf.destroy();
   }

void Cutlass_Decryptor::write(const byte data[], u32bit len)
   {
   msg_buf.append(data, len);
   }

void Cutlass_Decryptor::end_msg()
   {
   // Minimum size: IV + sequence no + length + MAC
   if(msg_buf.size() < iv_len + /*4*/ + 2 + mac_len)
      throw Decoding_Error("Incoming packet is too short");

   InitializationVector iv(msg_buf, iv_len);
   SecureVector<byte> recv_mac(msg_buf + msg_buf.size() - mac_len, mac_len);

   cipher->set_iv(iv);

   crypto_layer.start_msg();

#if 0
   mac->write(iv.begin(), iv.length());
#endif

   crypto_layer.write(msg_buf + iv_len, msg_buf.size() - iv_len - mac_len);
   crypto_layer.end_msg();

   SecureVector<byte> mac =
      crypto_layer.read_all(crypto_layer.message_count()-1);
   SecureVector<byte> plain =
      crypto_layer.read_all(crypto_layer.message_count()-2);

   if(mac != recv_mac)
      throw Integrity_Failure("MAC validation failure");

   // FIXME: change offsets once header is changed
#if 0
   u32bit seq_no = make_u32bit(plain[0], plain[1], plain[2], plain[3]);
   if(bad_seq_no(seq_no))
      throw Integrity_Failure("Bad sequence number");
#endif

   u32bit plain_size = make_u16bit(plain[2], plain[3]);

   if(plain.size() < plain_size + 4)
      throw Decoding_Error("Plaintext is truncated");

#if 0
   send(plain, plain_size); /* ignore any padding at the end */
#else
   /* Skip the length field */
   send(plain[0]);
   send(plain[1]);
   send(plain + 4, plain_size - 4);
#endif
   }

void Cutlass_Kex_Writer::start_msg()
   {
   msg_buf.destroy();
   }

void Cutlass_Kex_Writer::write(const byte buf[], u32bit length)
   {
   msg_buf.append(buf, length);
   }

void Cutlass_Kex_Writer::end_msg()
   {
   const u16bit buf_size = msg_buf.size();

   if(buf_size < 2)
      throw Invalid_Argument("Cutlass_Kex_Writer: Kex packet too short");

   const u16bit msg_size = buf_size - 2;

   send(msg_buf[0]);
   send(msg_buf[1]);
   send(get_byte(0, msg_size));
   send(get_byte(1, msg_size));
   send(msg_buf + 2, msg_size);
   }

void Cutlass_Kex_Reader::start_msg()
   {
   msg_buf.destroy();
   }

void Cutlass_Kex_Reader::write(const byte buf[], u32bit length)
   {
   msg_buf.append(buf, length);
   }

void Cutlass_Kex_Reader::end_msg()
   {
   const u16bit buf_size = msg_buf.size();

   if(buf_size < 4)
      throw Decoding_Error("Cutlass_Kex_Reader: Kex packet too short");

   u16bit length = make_u16bit(msg_buf[2], msg_buf[3]);
   if(length != msg_buf.size() - 4)
      throw Decoding_Error("Cutlass_Kex_Reader: Bad length field");

   send(msg_buf[0]);
   send(msg_buf[1]);
   send(msg_buf + 4, length);
   }
