/*
 * listener.c
 *
 * Listener initialization and loop code
 *
 * Copyright (c) 2004 Todd MacDermid <tmacd@synacklabs.net>
 *                    Jack Lloyd <lloyd@randombit.net>
 *                    nash e.f. <nash@solace.net>
 *
 */

#include <cutlass.h>

#include <inttypes.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/select.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

/*
 * cutlass_process_inbound will take the action needed by the inbound
 * packet. This includes updating key exchange, inputting packet buffers,
 * what have you. Highly dependent upon the connection state and the
 * packet contents. This is really a traffic-cop function, routing
 * to the appropriate logic module.
 */

int
cutlass_process_inbound(cutlass_t *cut_handle, struct cutlass_connection *conn,
			uint8_t *incoming_packet, int packet_size)
{
  struct timespec now;
  struct cutlass_packet_hdr packet_info;
  uint8_t *clear_packet = NULL;
  uint8_t packet_type;

  int retval;

  retval = 0;

  cutlass_time(&now);

  cutlass_sysmsg(cut_handle, CUT_SUPAA_DEBUG,
                 "Incoming packet of size %d\n", packet_size);

  switch(conn->conn_state)
  {
    case(CUTSTATE_DELETING):
      /* Do nothing, we're already on our way out. */
      break;
    case(CUTSTATE_UNINITIALZED):
    case(CUTSTATE_C_SENT_HELLO):
    case(CUTSTATE_S_SENT_HELLO):
    case(CUTSTATE_C_SENT_DH):
      cutlass_hdr_unpack( incoming_packet,
                          &packet_info.cut_type,
                          &packet_info.channel_id,
                          &packet_info.length);
      
      clear_packet = incoming_packet + CUT_HDR_LEN;
      retval = cutlass_process_handshake(cut_handle,
                                         conn,
                                         clear_packet,
                                         &packet_info,
                                         &now);
      if(retval < 0) {
        goto fail;
      }
      memcpy(&(conn->last_packet_sent), &now, sizeof(struct timespec));
      if(retval > 0) {
	return(0);
      }
      break;

    case(CUTSTATE_S_SENT_DH):
      clear_packet = cutlass_recv_process(conn,
                                          &packet_info,
                                          incoming_packet,
                                          packet_size);
      if(clear_packet == NULL) {
        cutlass_sysmsg( cut_handle,
                        CUT_DEBUG,
                        "cutlass_process_inbound: "
                        "S-SH state decryption failed\n");
        goto fail;
      } else {
        cutlass_conn_activate(cut_handle, conn, &now);

        retval = listener_handle_packet_types(cut_handle, conn, clear_packet,
                                              &packet_info, &now);        
        free(clear_packet); 
      }
      break;
    case(CUTSTATE_ACTIVE):
      clear_packet = cutlass_recv_process(conn,
                                          &packet_info,
                                          incoming_packet,
                                          packet_size);
      if(clear_packet == NULL)
      {
        cutlass_sysmsg( cut_handle,
                        CUT_INFO,
                        "cutlass_process_inbound: "
                        "Error decrypting packet\n");
        goto fail;
      }

      ////
      //// Thar be dragons...
      retval = listener_handle_packet_types(cut_handle, conn, clear_packet,
					    &packet_info, &now);

      // free and clear! it's like Sprint PCS... or something.
      free(clear_packet);
      break;

    case(CUTSTATE_CLOSE_SENT):
      {
        clear_packet = cutlass_recv_process(conn,
                                            &packet_info,
                                            incoming_packet,
                                            packet_size);
        if(clear_packet == NULL)
          goto fail;
        packet_type = packet_info.cut_type;
        if(packet_type == CUT_END_ACK) {
          cutlass_conn_del(cut_handle, conn);
	  free(clear_packet);
          return(0);
        }
        free(clear_packet);
        break;
      }
    case(CUTSTATE_CLOSE_ACKED):
      {
        clear_packet = cutlass_recv_process(conn,
                                            &packet_info,
                                            incoming_packet,
                                            packet_size);
        if(clear_packet == NULL)
          goto fail;
        packet_type = packet_info.cut_type;

        if(packet_type == CUT_END)
          cutlass_close_ack(cut_handle, conn);
        free(clear_packet);
        break;
      }
  }
  memcpy(&(conn->last_packet_received), &now, sizeof(struct timespec));

  if(retval < 0) {
    cutlass_conn_del(cut_handle, conn);
  } else {
    pthread_mutex_unlock(&(conn->conn_mutex));
  }

  return(0);
 fail:
  cutlass_sysmsg(cut_handle, CUT_DEBUG,
		 "process_inbound unlocking conn fail\n");
  pthread_mutex_unlock(&(conn->conn_mutex));
  return(-1);
}

/*
 * cutlass_get_inbound reads a packet into the packet buffer, sets the size
 * of the packet into the in_size variable, and returns a pointer to the
 * connection that packet belongs to. This will return NULL if no valid
 * connection is found. (New connections will be allocated where
 * appropriate, with their state set to CUTSTATE_UNINITIALZED).
 */


conn_t *
cutlass_get_inbound(cutlass_t *cut_handle, int fd,
		    uint8_t *packet, int *in_size)
{
  char err_message[CUT_MSG_LEN_MAX];
  socklen_t source_size;
  struct cutlass_connection *conn;
  struct cutlass_packet_hdr pkt_hdr;
  struct sockaddr_in hash_addr[2];
  struct sockaddr_in listen_addr;
  struct sockaddr_in source_addr;
  
  conn = NULL;
  
  if(fd == cut_handle->listen_socket) {
    memset(&listen_addr, 0, sizeof(listen_addr));
    listen_addr.sin_family = AF_INET;
    listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    listen_addr.sin_port = htons(
       (uint16_t)get_int_option(cut_handle->user_opts, "port")
       );
    memcpy(hash_addr, &listen_addr, sizeof(listen_addr));
    memset(&source_addr, 0, sizeof(source_addr));
    source_size = sizeof(source_addr);
    *in_size = recvfrom(fd, packet, CUT_MSG_LEN_MAX,
			0, (struct sockaddr *)&source_addr, &source_size);
    
    if(*in_size < CUT_HDR_LEN)
      {
	fprintf(stderr, "Truncated packet of size %d received\n", *in_size);
	return NULL;
      }
    
    memcpy(&(hash_addr[1]), &source_addr, source_size);
    
    if((conn = hashtable_addr_find(cut_handle, &source_addr)) == NULL) {
      cutlass_hdr_unpack(packet,
			 &pkt_hdr.cut_type,
			 &pkt_hdr.channel_id,
			 &pkt_hdr.length);
      if(pkt_hdr.cut_type == CUT_KEX_CHELLO) {
	conn = cutlass_conn_add_in(cut_handle, &source_addr);
      } else {
	cutlass_sysmsg(cut_handle, CUT_INFO,
		       "cutlass_get_inbound: Non-init packet received\n");
      }
    }
  } else {
    if ((conn = hashtable_fd_find(cut_handle, fd)) == NULL) {
      snprintf(err_message, CUT_MSG_LEN_MAX,
	       "cutlass_get_inbound: No conn for socket %d - %d bytes?\n",
	       fd, sizeof(fd));
      cutlass_sysmsg(cut_handle, CUT_ERROR, err_message);
    } else {
      *in_size = recv(fd, packet, CUT_MSG_LEN_MAX, 0);
    }
  }
  return(conn);
}


/*
 * cutlass_close_connection is called in response to either an EOF or
 * an error read from a socket. Time to shut either the connection
 * or the backconnect down.
 */

int
cutlass_close_connection(cutlass_t *cut_handle,
			 struct cutlass_connection *conn, int in_size)
{

  if(in_size == -1) {
    if(conn->conn_state == CUTSTATE_BACKCONNECT_SENT) {
      close(conn->socket);

    }
  } else {

  }
  return(0);
}

int
cutlass_fdset_add(cutlass_t *cut_handle, int fd)
{

  pthread_mutex_lock(&(cut_handle->fd_set_mutex));
  FD_SET(fd, &(cut_handle->saved_sockets));
  if(fd > cut_handle->max_fd) {
    cut_handle->max_fd = fd;
  }
  pthread_mutex_unlock(&(cut_handle->fd_set_mutex));
  return(0);
}

int
cutlass_fdset_del(cutlass_t *cut_handle, int fd)
{
  int i;
  int new_max;

  new_max = 0;

  pthread_mutex_lock(&(cut_handle->fd_set_mutex));
  FD_CLR(fd, &(cut_handle->saved_sockets));
  if(fd == cut_handle->max_fd) {
    for(i = 0; i < cut_handle->max_fd; i++) {
      if(FD_ISSET(i, &(cut_handle->saved_sockets))) {
	new_max = i;
      }
    }
    cut_handle->max_fd = new_max;
  }
  pthread_mutex_unlock(&(cut_handle->fd_set_mutex));
  return(0);
}


/*
 * cutlass_fdset_update updates and resets the list of file descriptors to 
 * select over before each loop. It does this by calling out to the 
 * listen_fd_remove_pop and listen_fd_add_pop functions above. After 
 * updating the saves copy, it copies it over to the "working" copy, which
 * is actually used in the select.
 */

int
cutlass_fdset_update(cutlass_t *cut_handle)
{

  pthread_mutex_lock(&(cut_handle->fd_set_mutex));
  
  memcpy(&(cut_handle->read_sockets), 
	 &(cut_handle->saved_sockets), 
	 sizeof(fd_set));
  
  memcpy(&(cut_handle->err_sockets), 
	 &(cut_handle->saved_sockets), 
	 sizeof(fd_set));
  
  pthread_mutex_unlock(&(cut_handle->fd_set_mutex));

  return(0);
}

/*
 * listener_thread is the loop that handles inbound traffic. Main loop for
 * the thread.
 */

void *
listener_thread(void *void_handle) {
  char err_message[CUT_MSG_LEN_MAX];
  char tmp_array[INET_ADDRSTRLEN];
  cutlass_t *cut_handle;
  int i;
  int in_size;
  int retval;
  uint8_t packet[CUT_MSG_LEN_MAX];
  struct cutlass_connection *conn;
  struct timeval timeout;

  pthread_detach(pthread_self());

  cut_handle = (cutlass_t *)void_handle;

  while(cut_handle->running_state) {
    cutlass_fdset_update(cut_handle);

    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    
    retval = select(((cut_handle->max_fd) + 1), 
		    &(cut_handle->read_sockets), 
		    NULL,
		    &(cut_handle->err_sockets), 
		    &timeout);
  
    if(retval < 0) {

      /* Do nothing, our file descriptor may have been closed from under us */
      /*      fprintf(stderr, "pselect returned error\n"); */
    } 

    if(retval > 0) {
       for(i = 0; i <= cut_handle->max_fd; i++) {
	if(FD_ISSET(i, &(cut_handle->read_sockets))) {
	  conn = cutlass_get_inbound(cut_handle, i, packet, &in_size);
	  if(conn != NULL) {
	    if(in_size <= 0) {
	      snprintf(err_message, CUT_MSG_LEN_MAX,
		       "listener thread - Error reading from %s\n",
		       cutlass_ntop(AF_INET, &(conn->remote_addr.sin_addr),
                                    tmp_array, INET_ADDRSTRLEN));
	      cutlass_sysmsg(cut_handle, CUT_ERROR, err_message);
	      cutlass_conn_del(cut_handle, conn);
	    } else {
	      cutlass_process_inbound(cut_handle, conn, packet, in_size);
	    }
	  }
	}
	if(FD_ISSET(i, &(cut_handle->err_sockets))) {
	  if(i != cut_handle->listen_socket) {
	    if((conn = hashtable_fd_find(cut_handle, i)) != NULL) {
	      snprintf(err_message, CUT_MSG_LEN_MAX,
		       "listener thread - Error reading from %s\n",
		       cutlass_ntop(AF_INET, &(conn->remote_addr.sin_addr),
				 tmp_array, INET_ADDRSTRLEN));
	      cutlass_sysmsg(cut_handle, CUT_ERROR, err_message);
	      cutlass_conn_del(cut_handle, conn);
	    }
	  }
	  snprintf(err_message, CUT_MSG_LEN_MAX, "error on fd %d\n", i);
	  cutlass_sysmsg(cut_handle, CUT_ERROR, err_message);
	}
      }
    }
  }
  pthread_exit(NULL);
}

/*
 * Takes the big nasty switch/case out of the inbound thingie.
 * Like I said above, thar be dragons here.
 *
 * FIX ME: I probably screwed up pointers right good and may not 
 * have the appropriate calling convention. Everyone should check
 * their calls below to make sure they're right.
 */
int
listener_handle_packet_types(cutlass_t *cut_handle,
			     conn_t *conn,
			     uint8_t *clear_packet,
			     struct cutlass_packet_hdr *packet_info,
			     struct timespec *now)
{
  int retval = 0;
  uint8_t packet_type;

  packet_type = packet_info->cut_type;

  cutlass_sysmsg(cut_handle, CUT_SUPAA_DEBUG, 
		 "Start of listener type switch\n");
  switch (packet_type) {
  case CUT_KEX_CHELLO:
  case CUT_KEX_SHELLO:
  case CUT_KEX_CDHG:
  case CUT_KEX_SDHG:
    cutlass_sysmsg(cut_handle, CUT_INFO,
                   "listener_handle_packet_types: "
                   "got a KEX packet on an active connection\n");
    return(1);
    break;
    
  case CUT_PING:
    retval = cutlass_pong(cut_handle, conn, clear_packet, now);
    break;

  case CUT_PONG:
    /* Do nothing */
    break;
    
  case CUT_INFO_REQ:
    retval = cutlass_info_send(cut_handle, conn, now);
    break;
    
  case CUT_CONN_INFO:
    retval = cutlass_info_parse(cut_handle, conn, clear_packet, now);
    break;
    
    ///
    /// TRANSPORT PACKET TYPES
    ///
    
  case CUT_TRANS_INIT:
    retval = cutlass_transport_init_parse(cut_handle, conn, clear_packet, 
                                          packet_info, now);
    break;
    
  case CUT_TRANS_INIT_ACK:
    retval = cutlass_transport_init_ack_parse(cut_handle, conn, clear_packet,
                                              packet_info, now);

  case CUT_TRANS_SEND:
    retval = cutlass_transport_parse(cut_handle, conn,
                                     clear_packet, packet_info, now);
    break;

  case CUT_TRANS_ACK:
    retval = cutlass_transport_ack_parse(cut_handle, conn, clear_packet, 
                                         packet_info, now);
    break;
    
  case CUT_AUDIO_INFO:
    /*    retval = cutlass_audio_info_parse(cut_handle, conn, clear_packet,
                                      packet_info, now);
    */
    break;
    
  case CUT_AUDIO_DATA:
    retval = cutlass_audio_data_parse(cut_handle, conn, clear_packet,
                                      packet_info, now);
    break;
    
  case CUT_AUDIO_REQ:
    retval = cutlass_audio_req_parse(cut_handle, conn, clear_packet,
                                     packet_info, now);
    break;
    
  case CUT_CHANNEL_CLOSE:
    retval = cutlass_channel_close_parse(cut_handle, conn, clear_packet,
                                         packet_info, now);
    break;
  case CUT_END:
    cutlass_close_ack(cut_handle, conn);
    break;
    
  default:
    cutlass_sysmsg(cut_handle, CUT_INFO,
                   "listener_handle_packet_types: "
                   "got an unknown packet type: %d\n", packet_type);
    return(1);
    break;
  }
  
  cutlass_sysmsg(cut_handle, CUT_SUPAA_DEBUG, "End of listener type switch\n");
  return retval;
}
